// ----------------------------------------------------------------------------
// Project:	Implementation Alternatives for Bottom-Up Evaluation
// Filename:	test7pull.cpp
// Purpose:	Time Measurements for Rule Implementations: Pull Method
// Last Change:	26.01.2010
// Language:	C++
// EMail:	brass@acm.org
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Univ. Halle, Von-Seckendorff-Platz 1, D-06120 Halle, GERMANY
// Copyright:	(c) 2010 by Stefan Brass
//		Downloads and modifications for tests and scientific purposes
//		permitted, provided that some acknowledgement of my work
//		remains, and modified versions are clearly marked as such.
//		There is no warranty at all - this code may contain bugs.
// ----------------------------------------------------------------------------

// Test Program 7:
//
// p(X, Z, 2) :- q(X, Y), r(Y, 5, Z).
// q(X, Y) :- s(X, Y), Y >= 0.
//
// r = {(i,j,k) | 1 <= i <= 10000, 1 <= j <= 100, 1 <= k <= 100}
//	r can be accessed with binding pattern bff (index on first argument)
// s = {(i,i) | 1 <= i <= 10000 }


#include <iostream>
using namespace std;

#if defined(_MSC_VER) || defined(_WINDOWS_)
#define WINDOWS
#else
#define UNIX
#endif

// ----------------------------------------------------------------------------
// Sizes for test:
// ----------------------------------------------------------------------------

const int S_TUPLES = 10000;

const int R_COL1_VALUES = 10000;

const int R_COL2_VALUES = 100;

const int R_COL3_VALUES = 100;

const bool PRINT_EXTENSION = false;

// ----------------------------------------------------------------------------
// Timing Functions (UNIX Version):
// ----------------------------------------------------------------------------

#ifdef UNIX
#include <sys/time.h>
#include <time.h>

class timer {
public:
	void start()
	{
		gettimeofday(&Time_Start, 0);
		Clock_Start = clock();
	}

	void end()
	{
		gettimeofday(&Time_End, 0);
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		long sec = Time_End.tv_sec - Time_Start.tv_sec;
		long usec = Time_End.tv_usec - Time_Start.tv_usec;
		if(usec < 0) {
			usec += 1000000;
			sec -= 1;
		}
		return sec * 1000 + (usec + 500) / 1000;
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	timeval Time_Start, Time_End;
	clock_t Clock_Start, Clock_End;
};
#endif

// ----------------------------------------------------------------------------
// Timing Functions (Windows Version):
// ----------------------------------------------------------------------------

#ifdef WINDOWS
#include <windows.h>
#include <time.h>

class timer {
public:
	void start()
	{
		GetSystemTimeAsFileTime(&(Start.ft));
		Clock_Start = clock();
	}

	void end()
	{
		GetSystemTimeAsFileTime(&(End.ft));
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		return static_cast<long> ((End.ns100 - Start.ns100) / 10000);
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	union {
		long long ns100;
		FILETIME ft;
	} Start, End;
	clock_t Clock_Start, Clock_End;
};
#endif


//-----------------------------------------------------------------------------
// Predicate s: Tuples generated by computation
//-----------------------------------------------------------------------------

class s_cursor {
public:
	s_cursor() { }

	void open() { i = 0; }

	bool fetch()
	{
		if(i < S_TUPLES) {
			i++;
			return true;
		}
		else
			return false;
	}

	int col_1() { return i; }

	int col_2() { return i; }

	void close() {}

private:
	int i;
};

//-----------------------------------------------------------------------------
// Predicate r (binding pattern bff): Tuples generated by computation
//-----------------------------------------------------------------------------

class r_cursor_bff {
public:
	r_cursor_bff() { }

	void open(int c1)
	{
		i = c1;
		if(i > 0 && i <= R_COL1_VALUES) {
			j = 0;
			k = R_COL3_VALUES;
		}
		else {
			j = R_COL2_VALUES + 1;
			k = 0;
		}
	}

	bool fetch()
	{
		if(j <= R_COL2_VALUES && k < R_COL3_VALUES) {
			k++;
			return true;
		}
		else if(j < R_COL2_VALUES) {
			j++;
			k = 1;
			return true;
		}
		else
			return false;
	}

	int col_1() { return i; }

	int col_2() { return j; }

	int col_3() { return k; }

	void close() {}

private:
	int i;
	int j;
	int k;
};


// ----------------------------------------------------------------------------
// Predicate q - Pull Method
// ----------------------------------------------------------------------------

class q_cursor {
	// q(X, Y) :- s(X, Y), Y >= 0.
public:
	void open()
	{
		s1.open();
	}

	bool fetch()
	{
		while(s1.fetch()) {
			if(s1.col_2() >= 0)
				return true;
		}
		return false;
	}

	int col_1() { return s1.col_1(); }
	int col_2() { return s1.col_2(); }

	void close()
	{
		s1.close();
	}
private:
	s_cursor s1;
};


// ----------------------------------------------------------------------------
// Predicate p - Pull Method
// ----------------------------------------------------------------------------


class p_cursor {
public:
    void open()
    {
        q1.open();
        q1_more = q1.fetch();
        if(q1_more)
            r1.open(q1.col_2());
    }
    bool fetch()
    {
        while(q1_more) {
            while(r1.fetch()) {
                if(r1.col_2() == 5)
                    return true;
            }
            r1.close();
            q1_more = q1.fetch();
            if(q1_more)
                r1.open(q1.col_2());
        }
        return false;
    }
    int col_1() { return q1.col_1(); }
    int col_2() { return r1.col_3(); }
    int col_3() { return 2; }
    void close()
    {
        q1.close();
        if(q1_more)
            r1.close();
    }
private:
    q_cursor q1;
    r_cursor_bff r1;
    bool q1_more;
};


//-----------------------------------------------------------------------------
// Main Program:
//-----------------------------------------------------------------------------

int main()
{
	timer t;
	int c1 = 0;
	int c2 = 0;
	int c3 = 0;

	cout << "Test of Method 2: Pull-Method (Program 7)\n";
	t.start();
	p_cursor p1;
	p1.open();
	while(p1.fetch()) {
		c1 += p1.col_1();
		c2 += p1.col_2();
		c3 += p1.col_3();
	}
	p1.close();
	t.end();

	if(PRINT_EXTENSION) {
		cout << "Extension of p:\n\n";
		p1.open();
		while(p1.fetch()) {
			cout << "p(" << p1.col_1() << ", " << p1.col_2()
				<< ", " << p1.col_3() << ").\n";
		}
		p1.close();
	}

	cout << "Column sums: col1=" << c1 << ", col2=" << c2 << ", col3="
		<< c3 << "\n\n";
	cout << "\nUsed time: " << t.realtime_msec() << " ms,  "
		<< "CPU: " << t.cputime_msec() << " ms.\n";

	return 0;
}

